-- Cryptarithmetic puzzle solver.
-- Brute force method.

module Main where

import Prelude hiding (lookup, (.))
-- import List ((\\), nub)

(.) f g x = f (g x)

(\\) :: (Eq a) => [a] -> [a] -> [a]
bs \\ cs =  flt bs cs

[]     `del` _             = []
(x:xs) `del` y | x == y    = xs
               | otherwise = x : xs `del` y

flt z []     =  z
flt z (x:xs) =  flt (del z x) xs

-- nub (meaning "essence") remove duplicate elements from its list argument.
nub :: (Eq a) => [a] -> [a]
nub l = nub' l []

nub' [] _     = []
nub' (x:xs) l = if x `elem` l then nub' xs l else x : nub' xs (x:l)

type VarName = Char
type CompiledPuzzle = [([VarName], VarName)]
type Mapping = [(VarName,Int)]

allChooseN :: (Eq a) => Int -> [a] -> [[a]]
allChooseN 0 _  = [[]]
allChooseN n [] = []
allChooseN n xs =
    foldr (++) [] (map (allChooseN' n xs) xs)

allChooseN' n xs x = (map (x:) (allChooseN (n-1) (xs\\[x])))

mappings :: [VarName] -> [Mapping]
mappings cs = map (zip cs) (allChooseN (length cs) [0..9])

lookup :: Mapping -> VarName -> Int
lookup m c = foldr (find c) (error "not bound") m

find c (c', v) v' = if (c == c') then v else v'

test :: CompiledPuzzle -> Mapping -> Bool
test puzzle m = testHlp m puzzle 0


testHlp m []          cy = cy == 0
testHlp m ((xs, x):r) cy =
    let
       cy'y = divMod (foldl (+) cy (map (lookup m) xs)) 10
       cy' = fst cy'y
       y = snd cy'y
    in
       lookup m x == y && testHlp m r cy'

compile puzzle = cHlp sum terms
                 where
                     puzzle' = map reverse (reverse puzzle)
                     terms = tail puzzle'
                     sum = head puzzle'

cHlp [] _      = []
cHlp (x:xs) ys = (concat (map (take 1) ys), x) :
                 cHlp xs (map (drop 1) ys)

solutions puzzle = (filter (test puzzle') (mappings vnames))
                   where
		       puzzle' = compile puzzle
                       vnames = nub (foldr1 (++) puzzle)

main = showSols (solutions puzzle) "\n"

-- puzzle = ["FOUR", "FIVE", "NINE"]
-- puzzle = ["SEND", "MORE", "MONEY"]
puzzle = ["ONE", "ONE", "ONE", "ONE", "ONE", "ONE", "SIX"]
-- puzzle = ["ETT", "TV", "TRE"]

showSols [] = showString "[]"
showSols (s:ss) = showChar '[' . showSol s . showSols' ss

showSols' [] = showChar ']'
showSols' (s:ss) = showChar ',' . showSol s . showSols' ss

showSol [] = showString "[]"
showSol (b:bs) = showChar '[' . showBind b . showSol' bs

showSol' [] = showChar ']'
showSol' (b:bs) = showChar ',' . showBind b . showSol' bs

showBind (x, v) = showString "(\'" . showChar x . showString "\'," . showString (show v) . showChar ')'
 
