module Prim(vfP) where

import Basic
import Val


vfP :: Name -> Val
vfP n | a == 0	  = Ctr n []
      | otherwise = Fun a (primFun n)
    where
        a = primArity n


primArity :: Name -> Int
primArity 'True'  = 0
primArity 'False' = 0
primArity 'Null'  = 0
primArity 'Cons'  = 2
primArity 'neg'   = 1
primArity '+ '    = 2
primArity '- '    = 2
primArity '* '    = 2
primArity '/ '    = 2
primArity 'mod'   = 2
primArity 'not'   = 1
primArity '< '    = 2
primArity '<='    = 2
primArity '=='    = 2
primArity '/='    = 2
primArity '>='    = 2
primArity '> '    = 2
primArity 'head'  = 1
primArity 'tail'  = 1
primArity 'null'  = 1
primArity 'if'    = 3
primArity _	  = error "Unknown primitive."


primFun 'Cons'  = primFun_Lambda1
primFun 'neg'   = primFun_Lambda2
primFun '+ '    = primFun_Lambda3
primFun '- '    = primFun_Lambda4
primFun '* '    = primFun_Lambda5
primFun '/ '    = primFun_Lambda6
primFun 'mod'   = primFun_Lambda7
primFun 'not'   = primFun_Lambda8
primFun '< '    = primFun_Lambda9
primFun '<='    = primFun_Lambda10
primFun '=='    = primFun_Lambda11
primFun '/='    = primFun_Lambda12
primFun '>='    = primFun_Lambda13
primFun '> '    = primFun_Lambda14
primFun 'head'  = primFun_Lambda15
primFun 'tail'  = primFun_Lambda16
primFun 'null'  = primFun_Lambda17
primFun 'if'    = primFun_Lambda18
primFun _       = error "Unknown primitive."


primFun_Lambda1  [hd, tl]  = Ctr 'Cons' [hd, tl]
primFun_Lambda2  [Int n]   = Int (-n)
primFun_Lambda3  [a, b]    = applyIII (+) a b
primFun_Lambda4  [a, b]    = applyIII (-) a b
primFun_Lambda5  [a, b]    = applyIII (*) a b
primFun_Lambda6  [a, b]    = applyIII div a b
primFun_Lambda7  [a, b]    = applyIII mod a b
primFun_Lambda8  [Ctr n _] = if n == 'True' then
                                 Ctr 'False' []
                             else
                                 Ctr 'True' []
primFun_Lambda9  [a, b]    = applyIIB (<)  a b
primFun_Lambda10 [a, b]    = applyIIB (<=) a b
primFun_Lambda11 [a, b]    = applyIIB (==) a b
primFun_Lambda12 [a, b]    = applyIIB (/=) a b
primFun_Lambda13 [a, b]    = applyIIB (>=) a b
primFun_Lambda14 [a, b]    = applyIIB (>)  a b
primFun_Lambda15 [Ctr n fs] = if n == 'Null' then
                                  error "Head of empty list"
                              else
				  fs !! 0 
primFun_Lambda16 [Ctr n fs] = if n == 'Null' then
                                  error "Tail of empty list"
                              else
				  fs !! 1 
primFun_Lambda17 [Ctr n _] = if n == 'Null' then
                                 Ctr 'True' []
                             else
                                 Ctr 'False' []
primFun_Lambda18 [Ctr n _, a, b] = if n == 'True' then a else b


applyIII op (Int n1) (Int n2) = Int (op n1 n2)


applyIIB op (Int n1) (Int n2) = if op n1 n2 then
				    Ctr 'True' []
				else
				    Ctr 'False' []
