! ****************************************************************************
! *                 F R E J A   R U N T I M E   S Y S T E M                  *
! *									     *
! *	Name:		RTS GC						     *
! *	File name:	rts_gc.s				             *
! *	Purpose:	Handwritten part of the garbage collector.     	     *
! *			To be included from the generated part of the	     *
! *			runtime system.					     *
! *	Programmer:	Henrik Nilsson					     *
! *	Version:	1.1						     *
! *	Date:		1999/02/24					     *
! *									     *
! ****************************************************************************

! ****************************************************************************
! Global state variables.
! ****************************************************************************
	
	.section ".data"
	.align	4
.enlarge_heap:
	.word	0

	
! ****************************************************************************
! Layout of the garbage collector's stack frame.
! ****************************************************************************

! VRSA_SIZE and FPO_VRSA is defined in "rts_defs.h".

! Other local variables here:
!	#define FOO_SIZE 4
!	FPO_FOO = (FPO_VRSA - FOO_SIZE)

! "Enumerate roots state". (Really needs only 4 bytes. 4 bytes padding.)
! void *ERS
! ERS0 = NULL
	
#define ERS_SIZE 8
#define ERS0 0
	FPO_ERS = FPO_VRSA - ERS_SIZE

! Root Pointer Vector. Must be large enough to hold at least 1000 pointers.
! graph *rpv[RPVSIZE]
	
#define RPV_SIZE 16384 * 4
	FPO_RPV = FPO_ERS - RPV_SIZE
	
! Reserve space for any C-arguments in addition to the 6 first which are
! passed in %o0 - %o5 and for which a 6 word argument save area is routinely
! reserved at FP + 68. Note: EXTRA_ARG_SIZE must be a multiple of 8!
! Currently, C-functions with up to 20 arguments (rather, up to 20 words of
! arguments) can be called.

#define EXTRA_ARGS_SIZE (14 * 4)
	
	gc_frame_size = VRSA_SIZE + ERS_SIZE + RPV_SIZE + EXTRA_ARGS_SIZE + 96

	
! ****************************************************************************
! Initialization.
! ****************************************************************************

! Note:	The heap base register (HB) is not really used currently. But if the
! GC strategy is changed, it may becomde important to quickly decide
! whether a pointer points into the young or an old generation. (Both during
! GC and during reduction (updates).)
	
	.global $RTS_initHeap
$RTS_initHeap:
	set	current_heap, %g1		! Load HB & HP.
	ld	[%g1 + 0], HB
	mov	HB, HP
	ld	[%g1 + 4], HL			! Load HL.
	add	HB, HL, HL
	set	.CAF_count * 12, %g1		! Subtract a safety margin. 
	sub	HL, %g1, HL
	set	.enlarge_heap, %g1		! Clear enlarge flag. 
	retl
	clr	[%g1]


! ****************************************************************************
! Garbage collector.
! ****************************************************************************
	
! Called when there is insufficient space on the heap.
! When GC completed, checks space again. If still not enough, an error
! is given and the program terminates.
!
! On entry:
! %o0		Required space in bytes.
!
! On exit:
! HB points to the new heap base.
! HP points to a space of at least %o0 bytes.
! HL contains the limit of the current heap.
! %o0-%o3, %g1	Corrupted.
!
! Note:	Don't forget that $RTS_CCallPrelude does NOT preserve SCAN & FWDTAG.
! Thus be very careful if C-routines are called while these are in use.

	.global $RTS_performGC
$RTS_performGC:
#ifdef DEBUGGING_ENABLED
	! Check for SIGINT.
	set	sigint, %g1
	ld	[%g1], %g1
	cmp	%g1, 0
	beq	1f
	nop
	set	.STR_userInterrupt, %o0
	ba	$RTS_handleError
	nop

	.section ".rodata"
.STR_userInterrupt:	
	.asciz	"User interrupt."
	.section ".text"

1:
	
#endif
		
	set	-gc_frame_size, %g1
	save	SP, %g1, SP

#if GC_PROFILING
	! Conditional collection of GC statistics/profiling & diagnostic msgs.
	call	$RTS_CCallPrelude
	nop
	set	gc_diagnostics, %g1
	ld	[%g1], %g1
	cmp	%g1, 0
	beq	1f
	nop
	call	diagnostic_gc_start
	nop
1:	sethi	%hi(gc_timer), %o0
	call	start_timer
	or	%o0, %lo(gc_timer), %o0
	call	$RTS_CCallPostlude
	nop
#endif
	
.once_more:				! Jump here if we must try once more.
	
	! Clear the mark vectors.
	! (RSG ensures they contain at least one word.)
	set	.CAF_marks, %o0
	set	.marker_marks_end - 4, %o1
.clear_marks:
	clr	[%o0]
	cmp	%o0, %o1
	bne	.clear_marks
	add	%o0, 4, %o0

	! Enlarge heap?
	set	.enlarge_heap, %g1
	ld	[%g1], %g1
	tst	%g1
	beq	1f
	nop

	call	$RTS_CCallPrelude
	nop
	set	other_heap, %o0		! Pointer to segment descriptor to %o0.
	ld	[%o0 + 4], %o1		! Current size to %o1.
	call	reallocate_segment	! Enlarge the segment.
	sll	%o1, 1, %o1		! Request twice the size.
	call	$RTS_CCallPostlude
	nop
	
1:	! Swap semi heaps, initialize GC state.
	set	current_heap, %o0
	set	other_heap, %o1
	ld	[%o0 + 0], HB		! Swap heap bases and load HB.
	swap	[%o1 + 0], HB
	st	HB, [%o0]
	mov	HB, HP			! Initialize HP, SCAN
	mov	HB, SCAN
	ld	[%o0 + 4], HL		! Swap sizes and load HL.
	swap	[%o1 + 4], HL
	st	HL, [%o0 + 4]
	add	HB, HL, HL
	set	.CAF_count * 12, %g1	! Subtract a safety margin. 
	sub	HL, %g1, HL
	add	TB, $STO_FWD, FWDTAG

	! Move objects referred from the spine stack.
	set	stacks, %g1		! Load spine stack base to %l1.
	ld	[%g1 + 0], %l0
	ld	[%g1 + 4], %l1
	add	%l0, %l1, %l1
	cmp	%i5, %l1		! Stack empty?
	beq	.out_move_loop		! Nothing to move if so.
	mov	%i5, %l0		! Top of spine stack to %l0.
	ld	[%l0], %o0		! Move element at %l0 otherwise.
.move_loop:
	ld	[%o0 + 0], %o1
	ld	[%o1 + DTO_MOVE], %g1
	jmpl	%g1, LINK
	nop
	st	%o0, [%l0]
	cmp	SCAN, HP		! Anything to relocate?
	beq	.out_reloc_loop		! If no, continue moving.
	add	%l0, 4, %l0		! Advance %l0 to next object.
	ld	[SCAN], %o1
.reloc_loop:				! Relocate objetcs.
	ld	[%o1 + DTO_RELOCATE], %g1
	jmpl	%g1, LINK
	nop
	cmp	SCAN, HP
	bne,a	.reloc_loop
	ld	[SCAN + 0], %o1
.out_reloc_loop:			
	cmp	%l0, %l1		! Entire stack traversed?
	bne,a	.move_loop		! Continue moving if not.
	ld	[%l0], %o0
.out_move_loop:			
	
#ifdef DEBUGGING_ENABLED
	! Move objects referred from the EDT.
	! Note:	Placing the EDT garbage collection here means that the CAFs
	! are not taken into account when the extra heap consumption due to
	! the EDT is computed.
	! Dont forget: $RTS_CCallPrelude does not preserve FWDTAG/SCAN.

	set	edt_heap_consumption, %g1	! Measure EDT heap consumption.
	st	HP, [%g1]
	set	ERS0, %g1		! Set state variable to initial state.
	st	%g1, [FP + FPO_ERS]
.edt_er_loop:
	call	$RTS_CCallPrelude
	std	FWDTAG, [FP + FPO_VRSA]
	set	FPO_RPV, %g1		! Load pointer to rpv to %o0.
	add	FP, %g1, %o0
	set	RPV_SIZE / 4, %o1	! Size of rpv to %o1
	call	tracer_enum_roots
	add	FP, FPO_ERS, %o2	! Load pointer to state variable to %o2
	mov	%o0, %l1		! Number of root references to %l1.
	call	$RTS_CCallPostlude
	ldd	[FP + FPO_VRSA], FWDTAG
	cmp	%l1, 0			! No (further) root references?
	beq	.out_edt_er_loop	! Jump out of loop if so.
	nop
	set	FPO_RPV, %g1		! Load rpv base to %l0.
	add	FP, %g1, %l0
	ld	[%l0], %l2		! Move obect referred from [%l0].
.move_loop2:
	ld	[%l2], %o0
	ld	[%o0 + 0], %o1
	ld	[%o1 + DTO_MOVE], %g1
	jmpl	%g1, LINK
	nop
	st	%o0, [%l2]		! Write back updated location of obj.
	cmp	SCAN, HP		! Anything to relocate?
	beq	.out_reloc_loop2	! If no, continue moving.
	add	%l0, 4, %l0		! Advance %l0 to next root reference.
	ld	[SCAN], %o1
.reloc_loop2:				! Relocate objetcs.
	ld	[%o1 + DTO_RELOCATE], %g1
	jmpl	%g1, LINK
	nop
	cmp	SCAN, HP
	bne,a	.reloc_loop2
	ld	[SCAN + 0], %o1
.out_reloc_loop2:			
	subcc	%l1, 1, %l1		! All root references traversed?
	bne,a	.move_loop2		! Continue moving if not.
	ld	[%l0], %l2
	ba	.edt_er_loop		! Enumerate further roots.
	nop
.out_edt_er_loop:	
#endif	
	
	! Move marked CAFs.
	set	.CAF_marks, %l0		! %l0 ptr to current word in mark vec.
	set	.CAF_marks_end, %l1	! %l1 ptr to end of mark vector.
	set	.CAF_table, %l2		! %l2 ptr to start of block of 32 CAFs.
.move_CAFs:
	mov	%l2, %l3		! %l3 ptr to current CAF.
	ld	[%l0], %l4		! Load current marks (32 mark flags).
	cmp	%l4, 0			! Any of these CAFs marked?
	beq	.out_move_CAFs2		! If not, advance to the next 32 CAFs.
	add	%l0, 4, %l0		! Advance %l0 to next word in mark vec.
.move_CAFs2:	
	btst	1, %l4			! Is CAF marked?
	beq	.out_reloc_loop3	! If not, check next mark flag.
	srl	%l4, 1, %l4		! Shift next mark flag into position.
	ld	[%l3 + 0], %o1		! Move this CAF.
	ld	[%o1 + DTO_MOVE], %g1
	jmpl	%g1, LINK
	mov	%l3, %o0
	cmp	SCAN, HP		! Any objects to relocate?
	beq	.out_reloc_loop3	! If not, check next mark flag.
	nop
	ld	[SCAN + 0], %o1
.reloc_loop3:				! Relocate objects.
	ld	[%o1 + DTO_RELOCATE], %g1
	jmpl	%g1, LINK
	nop
	cmp	SCAN, HP
	bne,a	.reloc_loop3
	ld	[SCAN + 0], %o1
.out_reloc_loop3:
	cmp	%l4, 0			! Any more marks among these 32?
	bne	.move_CAFs2		! If so, continue.
	add	%l3, 12, %l3		! Advance %l3 to next CAF.
.out_move_CAFs2:
	cmp	%l0, %l1		! End of mark vector?
	bne	.move_CAFs		! If not, continue.
	add	%l2, 32 * 12, %l2	! Advance %l2 to the next 32 CAFs.

	! Garbage collection of EDT goes here if we want an accurate measure
	! of the amount of extra space it requires.
	! Collect statistics here as well?

#ifdef DEBUGGING_ENABLED
	! Compute extra heap consumption due to the graphs retained by the
	! EDT. Nothe that ALL CAFs are included in this count. This is
	! pessimistic but a safe approximation.
	set	edt_heap_consumption, %g1
	ld	[%g1], %l0
	sub	HP, %l0, %l0
	st	%l0, [%g1]
#endif
		
	! Convert any forward pointers to indirection nodes in the CAF table.
	add	TB, $STO_IND, %g1		
	add	TB, $STO_FWD, FWDTAG		
	set	.CAF_table, %l0
	set	.CAF_table_end, %l1
	cmp	%l0, %l1
	beq	.out_fwd_to_ind_loop
	nop
	ld	[%l0], %l2
.fwd_to_ind_loop:
	cmp	%l2, FWDTAG		! Is this a forward pointer?
	beq,a	1f			! If so, convert to indirection node.
	st	%g1, [%l0]
1:	add	%l0, 12, %l0
	cmp	%l0, %l1
	bne	.fwd_to_ind_loop
	ld	[%l0], %l2
.out_fwd_to_ind_loop:

	! Enlarge other semi heap (i.e. other_heap again since we have
	! swapped them already).
	set	.enlarge_heap, %g1
	ld	[%g1], %l0
	tst	%l0
	beq	1f
	nop
	call	$RTS_CCallPrelude
	clr	[%g1]
	set	other_heap, %o0		! Pointer to segment descriptor to %o0.
	ld	[%o0 + 4], %o1		! Current size to %o1.
	call	reallocate_segment	! Enlarge the segment.
	sll	%o1, 1, %o1		! Request twice the size.
	call	$RTS_CCallPostlude
	nop

	! Enlarge heaps next time round if remaining size less than about
	! 25% (1/4) of entire size of a semi heap, or if the collection did
	! not free the requested amount of space.
1:	sub	HL, HP, %l0		! Did we get at least as many bytes
	cmp	%l0, %i0		! as requested?
	bl	1f			! No ...
	nop
	set	current_heap, %g1	! More left than 1/4 of current size?
	ld	[%g1 + 4], %l1		! (Prev. approx 5%, 1/16.)
	srl	%l1, 2, %l1
	cmp	%l0, %l1
	bg	2f
	nop
1:	set	.enlarge_heap, %g1	! Set .enlarge_heap.
	orn	%g0, 0, %l1
	st	%l1, [%g1]
	cmp	%l0, %i0		! Try once more if we could not satisfy
	bl	.once_more		! initial request.
	nop
2:

#if GC_PROFILING
	! Conclude  collection of GC statistics/profiling & diagnostic msgs.
	call	$RTS_CCallPrelude
	sub	HP, HB, %l0		! Amount of memory moved to %l0.
	sethi	%hi(gc_timer), %o0
	call	stop_timer
	or	%o0, %lo(gc_timer), %o0
	set	gc_diagnostics, %g1
	ld	[%g1], %g1
	cmp	%g1, 0
	beq	1f
	nop
	call	diagnostic_gc_end
	nop
1:	set	gc_log_file, %g1
	ld	[%g1], %g1
	cmp	%g1, 0			! Really comparison with NULL.
	beq	2f
	nop
	call	write_gc_log_entry	! Pass amount of memory moved as arg.
	mov	%l0, %o0		! Type size_t is assumed to be 32 bits.
2:	call	$RTS_CCallPostlude
	nop
#endif
	
	ret
	restore

	
! ****************************************************************************
! Special garbage collection routines.
! ****************************************************************************
		
	.global	$GC_trivialMove
$GC_trivialMove:
	retl
	nop


	.global	$GC_noMove	
$GC_noMove:
	set	.STR_no_move, %o0
	mov	%o1, %o2
	ld	[%o1 + DTO_INFO], %o3
	call	rts_internal_error	! Does not return.
	ld	[%o3], %o3

	.section ".rodata"
.STR_no_move:
	.asciz "Invalid move code invoked. Tag = %d = 0x%p, obj_type = %d\n"
	.section ".text"

	
	.global	$GC_noRelocate	
$GC_noRelocate:
	set	.STR_no_reloc, %o0
	mov	%o1, %o2
	ld	[%o1 + DTO_INFO], %o3
	call	rts_internal_error	! Does not return.
	ld	[%o3], %o3

	.section ".rodata"
.STR_no_reloc:
	.asciz "Invalid relocation code invoked. Tag = %d = 0x%p, obj_type = %d\n"
	.section ".text"

	
! Move code for IND (indirection node).
	.global	$GC_move_IND
$GC_move_IND:
	save	SP, -96, SP		! Convert to forward pointer.
	st	FWDTAG, [%i0 + 0]
	ld	[%i0 + 4], %o0		! move the indirected object.
	ld	[%o0 + 0], %o1
	ld	[%o1 + DTO_MOVE], %o2
	jmpl	%o2, LINK
	nop
	st	%o0, [%i0 + 4]		! Write forwarding address.
	ret
	restore	%o0, 0, %o0		! Return new address.


! Move code for VAP (vector application).
	.global	$GC_move_VAP
$GC_move_VAP:
	ld	[%o0 + 4], %o2		! Load ptr. to functions's desc. tbl.
	st	FWDTAG, [%o0 + 0]	! Convert to forward node.
	st	HP, [%o0 + 4]
	st	%o1, [HP + 0]		! Move initial part of VAP-node,
	st	%o2, [HP + 4]		! i.e. tag and desc. tbl. ptr.
	st	%o0, [HP + 8]		! And a pointer to the old node.
	ld	[%o2 + DTO_ARITY], %o1	! Get arity of function.
	sll	%o1, 2, %o1		! Compute size of VAP-node.
	add	%o1, 8, %o1
	mov	HP, %o0			! Return pointer to moved object.
	retl
	add	HP, %o1, HP		! Increment HP by size of VAP-node.
	

! Relocation code for VAP (vector application).
	.global $GC_relocate_VAP
$GC_relocate_VAP:
	save	SP, -96, SP
	ld	[SCAN + 4], %o1		! Get ptr. to fun's descriptor table.
	ld	[%o1 + DTO_MOVE], %l7	! Call "mark" function. We know that
	jmpl	%l7, LINK		! the mark functions do not need %o0.
	ld	[%o1 + DTO_ARITY], %l2	! Get arity of function.
	ld	[SCAN + 8], %l0		! Get pointer to old object.
	set	8, %l1			! Initialize argument index.
1:	ld	[%l0 + %l1], %o0	! Get ptr. to an object (argument).
	ld	[%o0 + 0], %o1		! Get tag.
	ld	[%o1 + DTO_MOVE], %l7	! Move the object.
	jmpl	%l7, LINK
	nop
	st	%o0, [SCAN + %l1]	! Store ptr to moved obj. in moved VAP.
	subcc	%l2, 1, %l2		! Decrement the argument counter.
	bne	1b			! Once more unless all args. moved.
	add	%l1, 4, %l1		! Advance index to next object.
	ret
	restore	SCAN, %l1, SCAN		! Advance SCAN ptr to next obj.
	

! Move code for TAP (traced application node).
	.global	$GC_move_TAP
$GC_move_TAP:
	ld	[%o0 + 4], %o2
	ld	[%o0 + 8], %o3
	ld	[%o0 + 12], %o4
	ld	[%o0 + 16], %o5
	st	FWDTAG, [%o0 + 0]
	st	HP, [%o0 + 4]
	st	%o1, [HP + 0]
	st	%o2, [HP + 4]
	st	%o3, [HP + 8]
	st	%o4, [HP + 12]
	st	%o5, [HP + 16]
	mov	HP, %o0
	retl
	add	HP, 20, HP


! Relocation code for TAP (traced application node).
	.global	$GC_relocate_TAP
$GC_relocate_TAP:
	save	SP, -96, SP
	ld	[SCAN + 4], %o0
	ld	[%o0 + 0], %o1
	ld	[%o1 + DTO_MOVE], %l7
	jmpl	%l7, LINK
	ld	[SCAN + 8], %l0
	st	%o0, [SCAN + 4]
	mov	%l0, %o0
	ld	[%o0 + 0], %o1
	ld	[%o1 + DTO_MOVE], %l7
	jmpl	%l7, LINK
	nop
	st	%o0, [SCAN + 8]
	ret
	restore	SCAN, 20, SCAN

	
! Move code for ZAP.
! Node that a ZAP-node must be treated as if it is a three word node since
! it later will be updated with a node consisting of up to three words.
	.global	$GC_move_ZAP
$GC_move_ZAP:
	ld	[%o0 + 4], %o2
	st	FWDTAG, [%o0 + 0]
	st	HP, [%o0 + 4]
	st	%o1, [HP + 0]
	st	%o2, [HP + 4]
	mov	HP, %o0
	retl
	add	HP, 12, HP


! Relocation code for ZAP.
	.global $GC_relocate_ZAP
$GC_relocate_ZAP:
	save	SP, -96, SP
	ld	[SCAN + 4], %o1		! Get ptr. to fun's descriptor table.
	ld	[%o1 + DTO_MOVE], %l7	! Call "mark" function. We know that
	jmpl	%l7, LINK		! the mark functions do not need %o0.
	nop
	ret
	restore	SCAN, 12, SCAN		! Note:	Size 12 since redex!
	

! Move code for HOLE.
	.global	$GC_move_HOLE
$GC_move_HOLE:
	st	FWDTAG, [%o0 + 0]
	st	HP, [%o0 + 4]
	st	%o1, [HP + 0]
	mov	HP, %o0
	retl
	add	HP, 12, HP


! Move code for BVEC.
	.global	$GC_move_BVEC
$GC_move_BVEC:
	ld	[%o0 + 4], %o2
	st	FWDTAG, [%o0 + 0]
	st	HP, [%o0 + 4]
	st	%o1, [HP + 0]
	st	%o2, [HP + 4]
	cmp	%o2, 0
	beq	2f
	or	%g0, 8, %o1
	add	%o2, 8, %o2
	ld	[%o0 + %o1], %o3
1:	st	%o3, [HP + %o1]
	add	%o1, 4, %o1
	cmp	%o1, %o2
	blu,a	1b
	ld	[%o0 + %o1], %o3
2:	mov	HP, %o0
	retl
	add	HP, %o1, HP


! Relocation code for BVEC.
	.global	$GC_relocate_BVEC
$GC_relocate_BVEC:
	ld	[SCAN + 4], %o0		! Get size.
	add	%o0, 8 + 3, %o0		! Compute size of entire object.
	andn	%o0, 3, %o0
	retl
	add	SCAN, %o0, SCAN

	
! Move code for FWD.
	.global	$GC_move_FWD
$GC_move_FWD:
	retl
	ld	[%o0 + 4], %o0


! Move code for INT.
	.global	$GC_move_INT
$GC_move_INT:
	ld	[%o0 + 4], %o2		! Get integer value.
	cmp	%o2, min_small_int	! Is this a small integer?
	bl	1f			! No, too negative.
	cmp	%o2, max_small_int
	bg	1f			! No, too positive.
	sethi	%hi($SI_0), %o1
	or	%o1, %lo($SI_0), %o1	! Use one of the static integers.
	sll	%o2, 3, %o2
	add	%o1, %o2, %o1
	st	FWDTAG, [%o0 + 0]
	st	%o1, [%o0 + 4]
	retl
	mov	%o1, %o0
1:	st	FWDTAG, [%o0 + 0]	! Conventional move, but ...
	st	HP, [%o0 + 4]
	add	TB, $STO_DINT, %o1	! convert to DINT since we know that
	st	%o1, [HP + 0]		! the integer is not small.
	st	%o2, [HP + 4]
	mov	HP, %o0
	retl
	add	HP, 8, HP


! Optimized move code for chains of list cells (or similar). 
	.global $GC_move_N_2
$GC_move_N_2:
	st	FWDTAG, [%o0]		! Must write FWD early if circ. struct.
        ld      [%o0 + 4], %o2		! Read head.
        ld      [%o0 + 8], %o3		! Read tail.
        ld      [%o3], %o4		! Read tag of next cell.
        cmp     %o1, %o4		! Still same tag?
        beq     1f			! If so, run optimized move code.
        st      HP, [%o0 + 4]		! Complete forward node with new addr.
        st      %o1, [HP]		! Move tag.
        st      %o2, [HP + 4]		! Move head.
        st      %o3, [HP + 8]		! Move tail.
        mov     HP, %o0			! Return new address.
        retl
        add     HP, 12, HP		! Update heap pointer.
1:	mov     HP, %o5			! Initialize destination address.
2:      st      %o2, [%o5 + 4]		! Move head.
        st      FWDTAG, [%o3]		! Write FWD tag to next cell.
        mov     %o3, %o0		! Advance to next cell.
        add     %o5, 12, %o5
        ld      [%o0 + 4], %o2		! Read head.
        ld      [%o0 + 8], %o3		! Read tail.
        ld      [%o3], %o4		! Read tag of next cell.
        cmp     %o1, %o4		! Still same tag?
        beq     2b			! If so, one more time round ...
        st      %o5, [%o0 + 4]		! Complete fwd node with new addr.
        st      %o2, [%o5 + 4]		! Last cell. Move head.
        st      %o3, [%o5 + 8]		! Move tail.
        st      %o5, [HP + 8]		! Write address to last cell.
        st      %o1, [HP + 12]		! Write original tag.
        add     TB, $STO_RLC_N_2, %o1	! Load tag of "multi cell reloc." obj.
        st      %o1, [HP]		! Write it to first moved cell.
        mov     HP, %o0			! Return new address.
        retl
        add     %o5, 12, HP		! Update heap pointer.

	
! Optimized relocation code for a chain of list cells (or similar) encoded
! as a "multi cell relocation" object.
	.global $GC_relocate_N_2
$GC_relocate_N_2:
	save	SP, -96, SP
        ld      [SCAN + 8], %l0		! Read address to last cell.
        ld      [SCAN + 12], %l1	! Read original tag.
1:      st      %l1, [SCAN]		! Restore tag of cell.
        ld      [SCAN + 4], %o0		! Read head.
        ld      [%o0 + 0], %o1		! Move it.
	ld	[%o1 + DTO_MOVE], %l7
	jmpl	%l7, LINK
        nop
        st      %o0, [SCAN + 4]		! Write relocated head.
        add     SCAN, 12, SCAN		! Advance scan pointer.
        cmp     %l0, SCAN		! Is next cell last one?
        bne     1b			! If not, continue.
        st      SCAN, [SCAN - 4]	! Make tail point to next cell.
        st      %l1, [SCAN]		! Last cell. Restore tag.
        ld      [SCAN + 4], %o0		! Read head.
        ld      [%o0], %o1		! Move it.
	ld	[%o1 + DTO_MOVE], %l7
        jmpl	%l7, LINK
        ld      [SCAN + 8], %l0		! Read tail.
        st      %o0, [SCAN + 4]		! Write relocated head.
        ld      [%l0], %o1		! Move tail.
	ld	[%o1 + DTO_MOVE], %l7
        jmpl	%l7, LINK
        mov     %l0, %o0
        st      %o0, [SCAN + 8]		! Write relocated tail.
        ret
        restore SCAN, 12, SCAN		! Advance scan pointer to next object.
	

! $GC_relocate_N_M: Relocates object with N fields, the first M of which is
! pointer fields.

	.global	$GC_relocate_1_0
$GC_relocate_1_0:
	retl
	add	SCAN, 8, SCAN


	.global	$GC_relocate_2_0
$GC_relocate_2_0:
	retl
	add	SCAN, 12, SCAN


	.global	$GC_relocate_2_1
$GC_relocate_2_1:
	save	SP, -96, SP
	ld	[SCAN + 4], %o0
	ld	[%o0 + 0], %o1
	ld	[%o1 + DTO_MOVE], %l7
	jmpl	%l7, LINK
	nop
	st	%o0, [SCAN + 4]
	ret
	restore	SCAN, 12, SCAN
