-- Version: 1.2
-- Date:    1999/09/23

-- Prelude functions and constants whose signatures are present in the
-- built-in Prelude (FrejaBuiltin). Thus the interface generated from this
-- file is of no use.
--
-- minInt, maxInt, minChar, maxChar are simple constants for which no code
-- is needed: they are only present in the interface. Since they are defined
-- in the builtin Prelude, they are not needed here. (The interface
-- generated from this file is thrown away anyway.)
--
-- The following types and functions from the Prelude are not defined
-- here and must thus be implemented as primitives elsewhere.
--
-- ()		The unit type.
-- Int		Integers.
-- Double	Double precision floating point numbers.
-- Char		The caracter type.
-- []		The list type.
-- (,)		All tuple types.
-- ->		The function type.
--
-- showsPrec	Method for Show.
--
-- _vAddInt	Auxiliary numeric function.
-- _vSubInt	Auxiliary numeric function.
-- _vMulInt	Auxiliary numeric function.
--
-- _selTuple_n_m	All tuple selectors.
--
-- _updatePat	Non-leaky pattern bindings.
--
-- error	The error function.
--
-- _fatbar	Special function for debug compilation of pattern matching.
-- _Fail	Special constructor for debug compilation of pattern matching.


module Prelude where

{-# trusted #-}
{-# optimize 0 #-}	-- For example, must NOT eta reduce in this module!

import qualified BuiltinPrelude as P
import qualified PreludeAux (String)
import qualified PreludeList as PL ((++))
import qualified PreludeText (show)


infixr 9 .
infixr 3 &&
infixr 2 ||
infixr 0 $


data Void


data Bool = False | True	-- Derives minBool, maxBool, toBool, fromBool


data Ordering = LT | EQ | GT	-- Derives minOrdering, maxOrdering,
				-- toOrdering, fromOrdering


-- Class Eval.

seq :: Eval a => a -> b -> b
seq x y = P.seq x y


-- Class Eq.

(==), (/=)  :: Eq a => a -> a -> P.Bool
x == y = x P.== y
x /= y = x P./= y


_eqEnum, _neEnum :: Enum a => a -> a -> P.Bool
_eqEnum x y = P._eqEnum x y
_neEnum x y = P._neEnum x y


_eqInt, _neInt :: P.Int -> P.Int -> P.Bool
_eqInt x y = P._eqInt x y
_neInt x y = P._neInt x y


_eqDouble, _neDouble :: P.Double -> P.Double -> P.Bool
_eqDouble x y = P._eqDouble x y
_neDouble x y = P._neDouble x y


-- Class Ord.

compare :: Ord a => a -> a -> P.Ordering
compare x y = P.compare x y


_compareEnum :: Enum a => a -> a -> P.Ordering
_compareEnum x y = P._compareEnum x y


_compareInt :: P.Int -> P.Int -> P.Ordering
_compareInt x y = P._compareInt x y


_compareDouble :: P.Double -> P.Double -> P.Ordering
_compareDouble x y = P._compareDouble x y


(<), (<=), (>=), (>) :: Ord a => a -> a -> P.Bool
x < y = x P.< y
x <= y = x P.<= y
x >= y = x P.>= y
x > y = x P.> y


_ltEnum, _leEnum, _geEnum, _gtEnum :: Enum a => a -> a -> P.Bool
_ltEnum x y = P._ltEnum x  y
_leEnum x y = P._leEnum x  y
_geEnum x y = P._geEnum x  y
_gtEnum x y = P._gtEnum x  y


_ltInt, _leInt, _geInt, _gtInt :: P.Int -> P.Int -> P.Bool
_ltInt x y = P._ltInt x  y
_leInt x y = P._leInt x  y
_geInt x y = P._geInt x  y
_gtInt x y = P._gtInt x  y


_ltDouble, _leDouble, _geDouble, _gtDouble :: P.Double -> P.Double -> P.Bool
_ltDouble x y = P._ltDouble x  y
_leDouble x y = P._leDouble x  y
_geDouble x y = P._geDouble x  y
_gtDouble x y = P._gtDouble x  y


-- Auxiliary function for Ord.

max :: (Ord a) => a -> a -> a
max x y | x P.>= y    = x
        | P.otherwise = y


min :: (Ord a) => a -> a -> a
min x y | x P.< y     = x
        | P.otherwise = y


-- Class Enum.

-- enumFromThen and enumFromThenTo will not work properly for Integer if and
-- when introduced. The step size is limited by Int, and so are the start
-- value as well as the following value due to the use of _fromEnum.

enumFrom :: Enum a => a -> [a]				-- [m ..]
enumFrom m = m P.: P._nextStep m 1 enumFrom []


enumFromThen :: Enum a => a -> a -> [a]			-- [m, m' ..]
enumFromThen m m' = eftHlp m
    where
	eftHlp i = i P.: P._nextStep i s eftHlp []
	s = P._fromEnum m' P.- P._fromEnum m
 

enumFromTo :: Enum a => a -> a -> [a]			-- [m .. n]
enumFromTo m n | m P.> n     = []
	       | P.otherwise = eftHlp m
    where
	eftHlp i = i P.: P._nextStepTo i 1 n eftHlp []


enumFromThenTo :: Enum a => a -> a -> a -> [a]		-- [m, m' .. n]
enumFromThenTo m m' n | s P.>= 0 P.&& m P.> n P.|| s P.< 0 P.&& m P.< n = []
		      | P.otherwise = efttHlp m
    where
	efttHlp i = i P.: P._nextStepTo i s n efttHlp []
	s = P._fromEnum m' P.- P._fromEnum m


succ :: Enum a => a -> a
succ x = P._nextStep x 1 P.id (P.error "Prelude.succ: No successor.")


pred :: Enum a => a -> a
pred x = P._nextStep x (-1) P.id (P.error "Prelude.pred: No predecessor.")


_toEnum :: Enum a => a -> P.Int -> a
_toEnum x n = P._toEnum x n


_fromEnum :: Enum a => a -> P.Int
_fromEnum x = P._fromEnum x


_nextStep :: Enum a => a -> P.Int -> (a -> b) -> b -> b
_nextStep x s f d = P._nextStep x s f d


_nextStepTo :: Enum a => a -> P.Int -> a -> (a -> b) -> b -> b
_nextStepTo x y s f d = P._nextStepTo x y s f d


-- Class Num.

(+), (-), (*) :: Num a => a -> a -> a
x + y = x P.+ y
x - y = x P.- y
x * y = x P.* y


_addInt, _subInt, _mulInt :: P.Int -> P.Int -> P.Int
_addInt x y = P._addInt x y
_subInt x y = P._subInt x y
_mulInt x y = P._mulInt x y


_addDouble, _subDouble, _mulDouble :: P.Double -> P.Double -> P.Double
_addDouble x y = P._addDouble x y
_subDouble x y = P._subDouble x y
_mulDouble x y = P._mulDouble x y


(^) :: Num a => a -> P.Int -> a
x ^ y = x P.^ y


_iPowInt :: P.Int -> P.Int -> P.Int
_iPowInt x y = P._iPowInt x y


_iPowDouble :: P.Double -> P.Int -> P.Double
_iPowDouble x y = P._iPowDouble x y


negate, abs, signum :: Num a => a -> a
negate x = P.negate x
abs x 	 = P.abs x
signum x = P.signum x


_negateInt, _absInt, _signumInt :: P.Int -> P.Int
_negateInt x = P._negateInt x
_absInt x    = P._absInt x
_signumInt x = P._signumInt x


_negateDouble, _absDouble, _signumDouble :: P.Double -> P.Double
_negateDouble x = P._negateDouble x
_absDouble x    = P._absDouble x
_signumDouble x = P._signumDouble x


intToNum :: Num a => a -> P.Int -> a
intToNum x y = P.intToNum x y


_intToNumDouble :: P.Double -> P.Int -> P.Double
_intToNumDouble x y = P._intToNumDouble x y


doubleToNum :: Num a => a -> P.Double -> a
doubleToNum x y = P.doubleToNum x y


_doubleToNumInt :: P.Int -> P.Double -> P.Int
_doubleToNumInt x y = P._doubleToNumInt x y


-- Class Real.

toInt :: Real a => a -> P.Int
toInt x = P.toInt x


toDouble :: Real a => a -> P.Double
toDouble x = P.toDouble x


-- Class Integral.

quot, rem, div, mod :: Integral a => a -> a -> a
x `quot` y = x `P.quot` y
x `rem` y  = x `P.rem` y
x `div` y  = x `P.div` y
x `mod` y  = x `P.mod` y


_quotInt, _remInt, _divInt, _modInt :: P.Int -> P.Int -> P.Int
x `_quotInt` y = x `P._quotInt` y
x `_remInt` y  = x `P._remInt` y
x `_divInt` y  = x `P._divInt` y
x `_modInt` y  = x `P._modInt` y


quotRem, divMod :: Integral a => a -> a -> (a, a)
quotRem x y = P.quotRem x y
divMod x y  = P.divMod x y 


-- Should be implemented in a better way.
-- PUSH 1; EVAL; GETI; EVAL; PUSH 0; GETI; DIVI; DEIND [1]; PUSH 1; GETI;
-- PUSH 0; GETI; MODI; MKINT; MKINT; UPDPACK 4 $Tuple_2; POP 2; RETURN
--
-- Or proper integration into the compilation schemes.

_quotRemInt, _divModInt :: P.Int -> P.Int -> (P.Int, P.Int)
_quotRemInt x y = (x `P._quotInt` y, x `P._remInt` y)
_divModInt x y = (x `P._divInt` y, x `P._modInt` y)


even, odd :: Integral a => a -> P.Bool
even x = P.even x
odd x  = P.odd x


_evenInt, _oddInt :: P.Int -> P.Bool
_evenInt x = P._evenInt x
_oddInt x  = P._oddInt x


-- Class Floating.

(/) :: Floating a => a -> a -> a
x / y = x P./ y


_fDivDouble :: P.Double -> P.Double -> P.Double
x `_fDivDouble` y = x `P._fDivDouble` y


recip, exp, log, sqrt :: Floating a => a -> a
recip x = P.recip x
exp x 	= P.exp x
log x 	= P.log x
sqrt x 	= P.sqrt x


_recipDouble, _expDouble, _logDouble, _sqrtDouble :: P.Double -> P.Double
_recipDouble x = 1.0 `_fDivDouble` x
_expDouble x 	= P._expDouble x
_logDouble x 	= P._logDouble x
_sqrtDouble x 	= P._sqrtDouble x


(**), logBase :: Floating a => a -> a -> a
x ** y = x P.** y
logBase x y = P.logBase x y


_powDouble, _logBaseDouble :: P.Double -> P.Double -> P.Double
x `_powDouble` y = x `P._powDouble` y
_logBaseDouble x y = P._logBaseDouble x y


sin, cos, tan :: Floating a => a -> a
sin x = P.sin x
cos x = P.cos x
tan x = P.tan x


_sinDouble, _cosDouble, _tanDouble :: P.Double -> P.Double
_sinDouble x = P._sinDouble x
_cosDouble x = P._cosDouble x
_tanDouble x = P._tanDouble x


asin, acos, atan :: Floating a => a -> a
asin x = P.asin x
acos x = P.acos x
atan x = P.atan x


_asinDouble, _acosDouble, _atanDouble :: P.Double -> P.Double
_asinDouble x = P._asinDouble x
_acosDouble x = P._acosDouble x
_atanDouble x = P._atanDouble x


sinh, cosh, tanh :: Floating a => a -> a
sinh x = P.sinh x
cosh x = P.cosh x
tanh x = P.tanh x


_sinhDouble, _coshDouble, _tanhDouble :: P.Double -> P.Double
_sinhDouble x = P._sinhDouble x
_coshDouble x = P._coshDouble x
_tanhDouble x = P._tanhDouble x


asinh, acosh, atanh :: Floating a => a -> a
asinh x = P.asinh x
acosh x = P.acosh x
atanh x = P.atanh x


_asinhDouble, _acoshDouble, _atanhDouble :: P.Double -> P.Double
_asinhDouble x = P._asinhDouble x
_acoshDouble x = P._acoshDouble x
_atanhDouble x = P._atanhDouble x


-- Class RealFloat.

truncate, round, ceiling, floor :: RealFloat a => a -> P.Int
truncate x = P.truncate x
round x    = P.round x
ceiling x  = P.ceiling x
floor x    = P.floor x


_truncateDouble, _roundDouble, _ceilingDouble, _floorDouble ::P.Double -> P.Int
_truncateDouble x = P._truncateDouble x
_roundDouble x    = P._roundDouble x
_ceilingDouble x  = P._ceilingDouble x
_floorDouble x    = P._floorDouble x


-- Auxiliary numeric functions.

_intToDouble :: P.Int -> P.Double
_intToDouble x = P._intToDouble x


{-
-- Derived automatically from the declaration of Bool.

minBool = P.False
maxBool = P.True


toBool :: P.Int -> P.Bool
toBool = P._toEnum P.False


fromBool :: P.Bool -> P.Int
fromBool = P._fromEnum
-}


(&&), (||) :: P.Bool -> P.Bool -> P.Bool
x && y = x P.&& y
x || y = x P.|| y


not :: P.Bool -> P.Bool
not x = P.not x


-- minInt, maxInt, minChar, maxChar are defined in the builtin prelude.


toChar :: P.Int -> P.Char
toChar = P._toEnum '\0'


fromChar :: P.Char -> P.Int
fromChar = P._fromEnum

{-
-- Derived automatically from the declaration of Ordering.

minOrdering = P.LT
maxOrdering = P.GT


toOrdering :: P.Int -> P.Ordering
toOrdering = P._toEnum P.LT


fromOrdering :: P.Ordering -> P.Int
fromOrdering = P._fromEnum
-}


-- Miscellaneous functions.

id :: a -> a
id x = x


const :: a -> b -> a
const x _ = x


_constId :: a -> b -> b
_constId _ x = x

(.) :: (b -> c) -> (a -> b) -> a -> c
f . g = \x -> f (g x)


flip :: (a -> b -> c) -> b -> a -> c
flip f x y = f y x


($) :: (a -> b) -> a -> b
f $ x = f x


asTypeOf :: a -> a -> a
asTypeOf x _ = x


_errorPos :: PreludeAux.String -> P.Int -> PreludeAux.String -> a
_errorPos f l m =
    P.error ("\""
	     PL.++ f
	     PL.++ "\", line "
	     PL.++ PreludeText.show l
	     PL.++ ": "
	     PL.++ m)


undefined :: a
undefined = P.error "Undefined"


_undefinedPos :: PreludeAux.String -> P.Int -> a
_undefinedPos f l =
    P.error ("Undefined used in file \""
	     PL.++ f
	     PL.++ "\", line "
             PL.++ PreludeText.show l
	     PL.++ ".")
