-- Version: 1.2
-- Date:    1999/09/23

-- Auxiliary Prelude definitions.

module PreludeAux (
           String,
           Maybe(..),
	   Either(..),
           maybe,
	   either,
	   strict,
	   gcd,
	   lcm,
	   subtract,
	   (^^),
	   curry,
	   uncurry,
	   until
       ) where

{-# trusted #-}

-- String and ShowS type aliases.

type String = [Char]


-- The type Maybe.

data Maybe a = Nothing | Just a


maybe :: b -> (a -> b) -> Maybe a -> b
maybe n f Nothing  = n
maybe n f (Just x) = f x


-- The type Either.

data Either a b = Left a | Right b


either :: (a -> c) -> (b -> c) -> Either a b -> c
either f g (Left x)  = f x
either f g (Right y) = g y


-- Auxiliary Eval functions.

strict :: (Eval a) => (a -> b) -> a -> b
strict f x = x `seq` f x


-- Auxiliary numeric functions.

gcd :: (Integral a) => a -> a -> a
gcd x y | x == zero && y == zero = error "Prelude.gcd: gcd 0 0 is undefined"
        | otherwise              = gcd' (abs x) (abs y)
    where
        gcd' x y | y == zero = x
                 | otherwise = gcd' y (x `rem` y)
        zero = intToNum x 0


lcm :: (Integral a) => a -> a -> a
lcm x y | x == zero || y == zero = zero
        | otherwise              = abs ((x `quot` (gcd x y)) * y)
    where
        zero = intToNum x 0


subtract :: (Num a) => a -> a -> a
subtract = flip (-)


(^^) :: (Floating a) => a -> Int -> a
x ^^ n = if n >= 0 then x^n else recip (x^(-n))


-- Miscellaneous functions.

curry :: ((a, b) -> c) -> a -> b -> c
curry f x y = f (x, y)


uncurry :: (a -> b -> c) -> ((a, b) -> c)
uncurry f p = f (fst p) (snd p)


until :: (a -> Bool) -> (a -> a) -> a -> a
until p f x | p x       = x
            | otherwise = until p f (f x)
