-- Version: 1.2
-- Date:    1999/09/23

-- Perhaps the VOps should be used when reading integers to avoid overflow.

module PreludeText (
           ShowS,
	   shows,
	   show,
	   showList,
           showChar,
	   showString,
           showParen,
	   readsIntBase,
	   readsInt,
	   readInt,
	   readsDouble,
	   readDouble,
	   _showsPrecArrow,
	   _showsPrecChar,
	   _showsPrecString,
	   _showsPrecList,
	   _showsPrecCon
       ) where

{-# trusted #-}

import PreludeAux (String, Maybe(..))
import PreludeList ((++), dropWhile, elem)

type ShowS = String -> String


shows :: (Show a) => a -> ShowS
shows = showsPrec 0


show :: (Show a) => a -> String
show x = shows x ""


showList :: (Show a) => [a] -> ShowS
showList []     = showString "[]"
showList (x:xs) = showChar '[' . shows x . showl xs
    where
	showl []     = showChar ']'
	showl (x:xs) = showString ", " . shows x . showl xs


showChar :: Char -> ShowS
showChar = (:)


showString :: String -> ShowS
showString = (++)


showParen :: Bool -> ShowS -> ShowS
showParen b p = if b then showChar '(' . p . showChar ')' else p


-- Read base b integer.
-- If the base is 0, the string itself determines the base according the
-- Freja syntax.


readsIntBase :: Int -> String -> Maybe (Int, String)
readsIntBase 0 x		     = (scanWhite >> scanInt) x
readsIntBase b x | 2 <= b && b <= 36 = (scanWhite >> (scanIntBase b)) x
readsIntBase b _		     =
    error ("PreludeText.readsIntBase: bad base " ++ show b)


readsInt :: String -> Maybe (Int, String)
readsInt = readsIntBase 10


readInt :: String -> Int
readInt x = case readsInt x of
		Nothing     -> error "PreludeText.readInt: no parse"
		Just (i, _) -> i


readsDouble :: String -> Maybe (Double, String)
readsDouble x = (scanWhite >> scanDouble) x


readDouble :: String -> Double
readDouble x = case readsDouble x of
		   Nothing     -> error "PreludeText.readDouble: no parse"
		   Just (i, _) -> i


-- Support

-- Scanner monad

type Scanner a = String -> Maybe (a, String)


m >>= k = \s -> case m s of
		       Nothing      -> Nothing
		       Just (x, s') -> k x s'
return x = \s -> Just (x, s)
m1 >> m2 = m1 >>= const m2
fail = \s -> Nothing


lit :: Char -> Scanner ()
lit c = \s -> case s of
	          []		      -> Nothing
		  (c':s') | c' == c   -> Just ((), s')
			  | otherwise -> Nothing


alt :: Scanner a -> Scanner a -> Scanner a
m1 `alt` m2 = \s -> case m1 s of
		        Nothing    -> m2 s
			r@(Just _) -> r


opt :: Scanner a -> Scanner (Maybe a)
opt m = \s -> case m s of
	          Nothing      -> Just (Nothing, s)
		  Just (x, s') -> Just (Just x, s')


optDef :: a -> Scanner a -> Scanner a
optDef x m = \s -> case m s of
	              Nothing      -> Just (x, s)
		      Just (y, s') -> Just (y, s')


scanWhite :: Scanner ()
scanWhite cs = Just ((), dropWhile isSpace cs)
    where
	isSpace c =c `elem` " \t\n\r\f\v\xA0"


scanInt :: Scanner Int
scanInt =
    scanSign	>>= \s ->
    scanUInt	>>= \i ->
    return (if s then -i else i)


scanUInt :: Scanner Int
scanUInt =
	(scanOctPfx		>>= \_ ->
	 scanPrimInt 8)
    `alt`
	(scanHexPfx		>>= \_ ->
	 scanPrimInt 16)
    `alt`
	(scanPrimInt 10		>>= \i  ->
	 optDef i (scanUICont i))


scanOctPfx = lit '0' >> (lit 'o' `alt` lit 'O')


scanHexPfx = lit '0' >> (lit 'x' `alt` lit 'X')


scanUICont b =
    lit '_' >> (if 2 <= b && b <= 36 then scanPrimInt b else fail)


scanIntBase :: Int -> Scanner Int
scanIntBase b =
    scanSign 		>>= \s ->
    scanPrimInt b	>>= \i ->
    return (if s then -i else i)
 

scanDouble :: Scanner Double
scanDouble =
    scanSign			>>= \s  ->
    scanIntPart			>>= \d1 ->
    optDef d1 (scanDecimals d1)	>>= \d2 ->
    optDef d2 (scanExponent d2)	>>= \d3 ->
    return (if s then -d3 else d3)


scanIntPart :: Scanner Double
scanIntPart []		          = Nothing
scanIntPart (c:cs) | isDecDigit c = sipHlp (ddToDouble c) cs
		   | otherwise    = Nothing
    where
        sipHlp d []			   = Just (d, [])
	sipHlp d ccs@(c:cs) | isDecDigit c = sipHlp (d*10.0+ddToDouble c) cs
			    | otherwise    = Just (d, ccs)


scanDecimals :: Double -> Scanner Double
scanDecimals d = lit '.' >> (scanDecimals1 d)


scanDecimals1 :: Double -> Scanner Double
scanDecimals1 d []		      = Nothing
scanDecimals1 d (c:cs) | isDecDigit c = sipHlp (d*10.0+ddToDouble c) 10.0 cs
		       | otherwise    = Nothing
    where
        sipHlp d s [] = Just (d / s, [])
	sipHlp d s ccs@(c:cs)
	    | isDecDigit c = sipHlp (d * 10.0 + ddToDouble c) (s * 10.0) cs
	    | otherwise    = Just (d / s, ccs)


scanExponent :: Double -> Scanner Double
scanExponent d =
    (lit 'e' `alt` lit 'E')	>>= \_ ->
    scanSign			>>= \s ->
    scanPrimInt 10		>>= \i -> 
    let
        e = min 400 i
	e1 = e `div` 100
	e2 = (e `div` 10) `mod` 10
	e3 = e `mod` 10
	d1 = d / 1.0e100^e1 / 1.0e10^e2 / 10.0^e3
	d2 = d * 1.0e100^e1 * 1.0e10^e2 * 10.0^e3
    in
        return (if s then d1 else d2)


scanSign :: Scanner Bool
scanSign =
    optDef False
	   ((lit '-' >> (return True))
	    `alt` (lit '+' >> (return False)))


scanPrimInt :: Int -> Scanner Int
scanPrimInt _ []		      = Nothing
scanPrimInt b (c:cs) | isExtDigit b c = spiHlp (edToInt c) cs
		     | otherwise      = Nothing
    where
        spiHlp i []			     = Just (i, [])
	spiHlp i ccs@(c:cs) | isExtDigit b c = spiHlp (i * b + edToInt c) cs
			    | otherwise      = Just (i, ccs)


isExtDigit :: Int -> Char -> Bool
isExtDigit b c = edToInt c < b


isDecDigit :: Char -> Bool
isDecDigit c = '0' <= c && c <= '9'


edToInt :: Char -> Int
edToInt c
    | '0' <= c && c <= '9' = fromChar c - fromChar '0'
    | 'A' <= c && c <= 'Z' = fromChar c - fromChar 'A' + 10
    | 'a' <= c && c <= 'z' = fromChar c - fromChar 'a' + 10
    | otherwise            = 1000


ddToDouble :: Char -> Double
ddToDouble c = toDouble (fromChar c - fromChar '0')


-- Internal function for showing values.
-- Note: none of these must refer to any CAFs, neither directly,
-- nor indirectly.

_showsPrecArrow d f = showString "<Function>"


_showsPrecChar d c = showChar '\'' . escapify [c] . showChar '\''


_showsPrecString d s = showChar '\"' . escapify s . showChar '\"'


_showsPrecList d xs = showList xs


_showsPrecCon d c = showString "<Con>"


escapify []            = id
escapify ('\NUL' : cs) = showString "\\NUL" . escapify cs
escapify ('\SOH' : cs) = showString "\\SOH" . escapify cs
escapify ('\STX' : cs) = showString "\\STX" . escapify cs
escapify ('\ETX' : cs) = showString "\\ETX" . escapify cs
escapify ('\EOT' : cs) = showString "\\EOT" . escapify cs
escapify ('\ENQ' : cs) = showString "\\ENQ" . escapify cs
escapify ('\ACK' : cs) = showString "\\ACK" . escapify cs
escapify ('\a'   : cs) = showString "\\a"   . escapify cs
escapify ('\b'   : cs) = showString "\\b"   . escapify cs
escapify ('\t'   : cs) = showString "\\t"   . escapify cs
escapify ('\n'   : cs) = showString "\\n"   . escapify cs
escapify ('\v'   : cs) = showString "\\v"   . escapify cs
escapify ('\f'   : cs) = showString "\\f"   . escapify cs
escapify ('\r'   : cs) = showString "\\r"   . escapify cs
escapify ('\SO'  : cs) = showString "\\SO"  . escapifyH cs
escapify ('\SI'  : cs) = showString "\\SI"  . escapify cs
escapify ('\DLE' : cs) = showString "\\DLE" . escapify cs
escapify ('\DC1' : cs) = showString "\\DC1" . escapify cs
escapify ('\DC2' : cs) = showString "\\DC2" . escapify cs
escapify ('\DC3' : cs) = showString "\\DC3" . escapify cs
escapify ('\DC4' : cs) = showString "\\DC4" . escapify cs
escapify ('\NAK' : cs) = showString "\\NAK" . escapify cs
escapify ('\SYN' : cs) = showString "\\SYN" . escapify cs
escapify ('\ETB' : cs) = showString "\\ETB" . escapify cs
escapify ('\CAN' : cs) = showString "\\CAN" . escapify cs
escapify ('\EM'  : cs) = showString "\\EM"  . escapify cs
escapify ('\SUB' : cs) = showString "\\SUB" . escapify cs
escapify ('\ESC' : cs) = showString "\\ESC" . escapify cs
escapify ('\FS'  : cs) = showString "\\FS"  . escapify cs
escapify ('\GS'  : cs) = showString "\\GS"  . escapify cs
escapify ('\RS'  : cs) = showString "\\RS"  . escapify cs
escapify ('\US'  : cs) = showString "\\US"  . escapify cs
escapify ('\DEL' : cs) = showString "\\DEL" . escapify cs
escapify ('\''   : cs) = showString "\\\'"  . escapify cs
escapify ('\"'   : cs) = showString "\\\""  . escapify cs
escapify ('\\'   : cs) = showString "\\\\"  . escapify cs
escapify (c : cs)
    | c < ' ' || c >= '\DEL' && c <= '\x9f' =
        showString "\\" . shows (fromChar c) . escapifyDigit cs
    | otherwise = showChar c . escapify cs


escapifyH cs@('H' : _) = showString "\\&" . escapify cs
escapifyH cs           = escapify cs


escapifyDigit cs@(c: _) | '0' <= c && c <= '9' = showString "\\&" . escapify cs
escapifyDigit cs           		       = escapify cs


